/************************************************
* CEH.C - Critical Error Handling functions		*
*												*
* 900424 MCMason - updated for Turbo C v2.0		*
* 8911?? RHShaw  - originally written			*
************************************************/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <conio.h>
#include <dos.h>

#define TRUE	(1==1)
#define FALSE	(!TRUE)
#define isin(a,b,c) (((a)<=(c)) && ((c)<=(b)))

#define _CRITERR_FAIL   3	/* fail on error	*/
#define _CRITERR_ABORT  2	/* abort on error	*/
#define _CRITERR_RETRY  1	/* retry on error	*/
#define _CRITERR_IGNORE 0	/* ignore error		*/

static char _bar[]="============================"
	"=======================================\n\r";

/************************************************
* cehDevErrMsg - return ptr to the appropriate	*
* error message indicated by the device error	*
* code.											*
*												*
* INP:	DevErrCode - the device error code		*
* OUT:	ptr to static (READ ONLY!) message		*
************************************************/
static const char *_cehDevErrMsgs[]=
    {
    "Write-protected disk",
    "Unknown unit",
    "Drive not ready",
    "Unknown command",
    "CRC error in data",
    "Bad request-structure length",
    "Seek error",
    "Unknown media type",
    "Sector not found",
    "Printer out of paper",
    "Write fault",
    "Read fault",
    "General failure",
	NULL,
	NULL,
	"Invalid disk change"
    };

const char *cehDevErrMsg(int DevErrCode)
	{
	const char *retval;
	retval = (const char *)NULL;
	if (isin(0,15,DevErrCode))
		retval = _cehDevErrMsgs[DevErrCode];
	if (retval == (const char *)NULL)
		retval = "** UNKNOWN ERROR CODE **";
	return retval;
	}

/************************************************
* cehDiskErrLoc - get string describing which	*
* area of the disk contained the error.			*
*												*
* INP:	DiskLocCode - the code that	specifies	*
*			the location of the disk error.		*
************************************************/
static const char *_cehDiskErrLocs[] =
    {
    "MS-DOS",
    "FAT",
    "Directory",
    "Data area"
    };

const char *cehDiskErrLoc(int DiskLocCode)
	{
	if (isin(0,3,DiskLocCode))
		return _cehDiskErrLocs[DiskLocCode];
	else
		return "** UNKNOWN DISK LOCATION **";
	}

/************************************************
* cehGetDevName - return the device name		*
*												*
* INP:	DevHdrPtr - ptr to device header		*
* OUT:	ptr to static buffer holding name		*
************************************************/
const char *cehGetDevName(unsigned far *DevHdrPtr)
	{
	static char x[20];
	if (DevHdrPtr[2]&0x8000)
		sprintf(x,"%8.8s",(char *)(DevHdrPtr+5));
	else
		sprintf(x,"Block %2d units",
							(DevHdrPtr[5]&0xff));
	return x;
	}

/************************************************
* cehPrintDevHeader - print some interesting	*
* statistics about the device that had a failure*
*												*
* INP:	DevHdrPtr - ptr to the device header	*
************************************************/
void cehPrintDevHeader(unsigned far *DevHdrPtr)
	{
	unsigned tmp;

	/* Get the device attribute word */
	tmp = DevHdrPtr[2];

	/* Print device name */
	cprintf("Device: %s\n\r",cehGetDevName(
									DevHdrPtr));

	if (tmp & 0x8000)
		{
		/* Print character-based device data */
		cprintf("Output until busy %ssupported"
				"\n\r",(tmp&0x2000)?"":"not ");
		if (tmp&0x02)
			cprintf("STDOUT driver\n\r");
		if (tmp&0x10) cprintf("CON driver\n\r");
		if (tmp&0x08) cprintf("CLOCK$ driver\n\r");
		if (tmp&0x04) cprintf("NUL driver\n\r");
		if (tmp&0x01) cprintf("STDIN driver\n\r");
		}
	else
		{
		/* Print block-based device data */
		cprintf("Use %s to determine media chara"
				"cteristics\n\r",(tmp&0x2000) ?
				"BIOS parameter block" :
				"Media ID byte");
		cprintf("IOCTL read & write %ssupported"
			"\n\r",(tmp&0x4000)?"":"not ");
		cprintf("Removable media %ssupported\n\r",
			(tmp&0x0800)?"":"not ");
		cprintf("Generic IOCTL & logical drives "
			"%ssupported\n\r",(tmp&0x40)?"":"not ");
		if (tmp&0x02)
			cprintf("Uses 32 bit sectors\n\r");
		}
	}

/************************************************
* CriticalErrorHandler - the replacement of the	*
* DOS critical error handler.					*
*												*
* INP:	errval - the error code from DOS		*
*		deverr - the device error code			*
*		devhdr - ptr to the device driver header*
* OUT:	_CRITERR_IGN - ignore the error			*
*		_CRITERR_RET - retry the operation		*
*		_CRITERR_ABT - abort the program		*
*		_CRITERR_FAIL - flag operation as failed*
* NOTE:	This function NOT declared as interrupt	*
*		handler function, because it is called	*
*		FROM a small handler that actually does	*
*		handle the interrupt.					*
************************************************/
int CriticalErrorHandler(unsigned errval,
		unsigned deverr, unsigned segment,
		unsigned offset)
    {
	unsigned	drive, loc, c, far *devhdr,
	            prev=FALSE, fail_allowed=FALSE,
	            attribute_word;
    char		*action, *comma_space = ", ",
	            *nullstr = "",buf[80];
    char far	*hdr;

	devhdr = MK_FP(segment,offset);

	/* Show the entry parameters */
	cprintf("\r\n");
	cprintf(_bar);
	cprintf("errval = %04x = %s\n\r",
			errval,cehDevErrMsg(errval));
	cprintf("deverr = %04x\n\r",deverr);
	cprintf("devhdr = %p\n\r",devhdr);
	
	/* Show the device characteristics */
	cprintf(_bar);
	cehPrintDevHeader(devhdr);

	/* Give a 'standard' error message */
	cprintf(_bar);
    if(deverr & 0x8000)
        {
        /*******************
        * Not a disk error *
        *******************/
        hdr = (char far *)devhdr;
        hdr += 4;
        attribute_word = *((unsigned far *)hdr);
        if(attribute_word & 0x8000)
            sprintf(buf,
                "Bad File Allocation Table\n\r");
        }
    else
        {
        /*******************
        * Disk drive error *
        *******************/
        action = (deverr&0x0100) ? "writing" : "reading";
        loc = ((deverr & 0x0600) >> 9);
        drive = (deverr & 0x00ff)+'A';
        sprintf(buf,
            "Error in %s, %s drive %c, %s\n\r",
            cehDiskErrLoc(loc),action,drive,
            cehDevErrMsg(errval & 0x00ff));
        }

    /* prompt user for disposition */
    while(TRUE)
        {
        cprintf(buf);
        if(deverr & 0x0800)
            {
            prev = TRUE;
            cprintf("Abort");
            if(_osmajor > 2)
                fail_allowed = TRUE;
            }
        if(deverr & 0x1000)
            {
            cprintf("%sRetry",
                (prev ? comma_space : nullstr));
            prev = TRUE;
            }
        if(deverr & 0x2000)
            {
            cprintf("%sIgnore",
                (prev ? comma_space : nullstr));
            prev = TRUE;
            }
        if(fail_allowed)
            cprintf("%sFail",
                (prev ? comma_space : nullstr));
        cprintf("? ");
        c = getche();
        cprintf("\n\r");
        switch(toupper(c))
            {
            case 'A':
				hardresume(_CRITERR_ABORT);
                break;
            case 'R':
				hardresume(_CRITERR_RETRY);
                break;
            case 'I':
				hardresume(_CRITERR_IGNORE);
                break;
            case 'F':
                if(fail_allowed)
					hardresume(_CRITERR_FAIL);
            }
        }
    }
